#pragma once
#include <stm32f4xx.h>
enum class jy61_device_communication_status_t : uint8_t
{
    PacketEnd,
    PacketBegin,
    PacketData
};
struct jy61_accel_data_raw_t
{
    int16_t accel_x_raw;
    int16_t accel_y_raw;
    int16_t accel_z_raw;
    int16_t temperature_raw;
};
struct jy61_gyro_data_raw_t
{
    int16_t gyro_x_raw;
    int16_t gyro_y_raw;
    int16_t gyro_z_raw;
    int16_t temperature_raw;
};
struct jy61_attitude_data_raw_t
{
    int16_t roll_raw;
    int16_t pitch_raw;
    int16_t yaw_raw;
    int16_t temperature_raw;
};
union jy61_packet {
    struct
    {
        [[maybe_unused]] uint8_t magic;
        uint8_t kind;
        union {
            jy61_accel_data_raw_t accel;
            jy61_gyro_data_raw_t gyro;
            jy61_attitude_data_raw_t attitude;
        };
        [[maybe_unused]] uint8_t checksums;
    };
    uint8_t data[11];
};
class jy61_device_t
{
  private:
    USART_TypeDef *usart;
    volatile jy61_device_communication_status_t communication_status;
    volatile jy61_packet buffer;
    volatile uint8_t current_received_size;
    void handle_packet() noexcept;

  public:
    volatile jy61_accel_data_raw_t accel_raw{};
    volatile jy61_gyro_data_raw_t gyro_raw{};
    volatile jy61_attitude_data_raw_t attitude_raw{};
    explicit jy61_device_t(USART_TypeDef *usart) noexcept;
    void begin() noexcept;
    void handle_isr() noexcept;
};