#include "../Inc/app_main.h"
#include "global.h"
#include "tim.h"
#include <ILI9341.h>
#include <cmath>
#include <encoder_reader_t.h>
#include <lvgl.h>
#include <ui.h>

// Frame buffers
static lv_disp_draw_buf_t disp_buf;
constexpr const size_t DISP_BUFF_SIZE = 320 * 240 / 10;
static lv_color_t buf_1[DISP_BUFF_SIZE];
static encoder_reader_t encoder = encoder_reader_t(EncoderA_GPIO_Port, EncoderA_Pin, EncoderB_GPIO_Port, EncoderB_Pin);

extern "C" void usart2_irq()
{
    // we do not call `HAL_UART_IRQHandler` for we manage the buffer totally,
    // so we cannot use some HAL function for this UART
    jy61_device.handle_isr();
}

extern "C" void app_pre_init()
{
    // do nothing
}
extern "C" void app_init()
{
    // do nothing
}
extern "C" void app_sys_init()
{
    // do nothing
}

static bool is_buzzer_on()
{
    return __HAL_TIM_GET_COMPARE(&htim2, TIM_CHANNEL_1) == 250;
}

static void set_buzzer(bool on)
{
    __HAL_TIM_SetCompare(&htim2, TIM_CHANNEL_1, on ? 250 : 500);
}

extern "C" void HAL_GPIO_EXTI_Callback(uint16_t GPIO_Pin)
{

    switch (GPIO_Pin)
    {
    case EncoderA_Pin:
        encoder.handle_a();
        break;
    case EncoderB_Pin:
        encoder.handle_b();
        break;
    default:
        break;
    }
}
extern "C" void HAL_TIM_PeriodElapsedCallback(TIM_HandleTypeDef *htim)
{
    if (htim->Instance == TIM4)
    {
        if (!fan_output)
        {
            set_buzzer(false);
            __HAL_TIM_SetCompare(&htim1, TIM_CHANNEL_1, 0);
            __HAL_TIM_SetCompare(&htim1, TIM_CHANNEL_3, 0);
            __HAL_TIM_SetCompare(&htim3, TIM_CHANNEL_1, 0);
            __HAL_TIM_SetCompare(&htim3, TIM_CHANNEL_3, 0);
            return;
        }

        float dt = 0.01;
        double t = (time_elapsed += dt);

        float max_control_value = 850;
        double pitch = (jy61_device.attitude_raw.pitch_raw - pitch_bias_raw) / 32768.0 * M_PI;
        double roll = (jy61_device.attitude_raw.roll_raw - roll_bias_raw) / 32768.0 * M_PI;

        if (t < 0)
        {
            set_buzzer(static_cast<int>(-t * 1000) % 1000 < 500);
            t = 0;
        }
        else if (is_buzzer_on())
        {
            set_buzzer(false);
        }

        float target_pitch = curve_pitch(t);
        float target_roll = curve_roll(t);

        float pitch_control = pid_pitch.compute((float)(target_pitch - pitch), dt);
        float roll_control = pid_roll.compute((float)(target_roll - roll), dt);

        if (pitch_control < 0)
        {
            pitch_control = std::min(-pitch_control, max_control_value);
            __HAL_TIM_SetCompare(&htim1, TIM_CHANNEL_1, 0);
            __HAL_TIM_SetCompare(&htim1, TIM_CHANNEL_3, pitch_control);
        }
        else
        {
            pitch_control = std::min(pitch_control, max_control_value);
            __HAL_TIM_SetCompare(&htim1, TIM_CHANNEL_1, pitch_control);
            __HAL_TIM_SetCompare(&htim1, TIM_CHANNEL_3, 0);
        }

        if (roll_control < 0)
        {
            roll_control = std::min(-roll_control, max_control_value);
            __HAL_TIM_SetCompare(&htim3, TIM_CHANNEL_1, 0);
            __HAL_TIM_SetCompare(&htim3, TIM_CHANNEL_3, roll_control);
        }
        else
        {
            roll_control = std::min(roll_control, max_control_value);
            __HAL_TIM_SetCompare(&htim3, TIM_CHANNEL_1, roll_control);
            __HAL_TIM_SetCompare(&htim3, TIM_CHANNEL_3, 0);
        }
    }
}

extern "C" void encoder_read([[maybe_unused]] lv_indev_drv_t *indev, lv_indev_data_t *data)
{
    data->enc_diff = encoder.count;
    encoder.count = 0;
    data->key = LV_KEY_ENTER;
    if (HAL_GPIO_ReadPin(EncoderKey_GPIO_Port, EncoderKey_Pin) == GPIO_PIN_RESET)
    {
        data->state = LV_INDEV_STATE_PRESSED;
    }
    else
    {
        data->state = LV_INDEV_STATE_RELEASED;
    }
}

extern "C" void app_main()
{
#if 0
    HAL_TIM_PWM_Start(&htim2, TIM_CHANNEL_1);
    HAL_TIM_PWM_Start(&htim1, TIM_CHANNEL_1);
    HAL_TIM_PWM_Start(&htim1, TIM_CHANNEL_2);
    HAL_TIM_PWM_Start(&htim1, TIM_CHANNEL_3);
    HAL_TIM_PWM_Start(&htim1, TIM_CHANNEL_4);
    HAL_TIM_PWM_Start(&htim3, TIM_CHANNEL_1);
    HAL_TIM_PWM_Start(&htim3, TIM_CHANNEL_2);
    HAL_TIM_PWM_Start(&htim3, TIM_CHANNEL_3);
    HAL_TIM_PWM_Start(&htim3, TIM_CHANNEL_4);
    __HAL_TIM_SetCompare(&htim1, TIM_CHANNEL_2, 0);
    __HAL_TIM_SetCompare(&htim1, TIM_CHANNEL_4, 0);
    __HAL_TIM_SetCompare(&htim3, TIM_CHANNEL_2, 0);
    __HAL_TIM_SetCompare(&htim3, TIM_CHANNEL_4, 0);
    __HAL_TIM_SetCompare(&htim1, TIM_CHANNEL_1, 0);
    __HAL_TIM_SetCompare(&htim1, TIM_CHANNEL_3, 0);
    __HAL_TIM_SetCompare(&htim3, TIM_CHANNEL_1, 0);
    __HAL_TIM_SetCompare(&htim3, TIM_CHANNEL_3, 0);
    set_buzzer(false);
    for(;;) {

    }
#endif
    pid_pitch.Kp = 100;
    pid_pitch.Ki = 0.3;
    pid_pitch.Kd = 1000;
    pid_roll.Kp = 100;
    pid_roll.Ki = 0.3;
    pid_roll.Kd = 1000;

    jy61_device.begin();
    HAL_TIM_PWM_Start(&htim1, TIM_CHANNEL_1);
    HAL_TIM_PWM_Start(&htim1, TIM_CHANNEL_2);
    HAL_TIM_PWM_Start(&htim1, TIM_CHANNEL_3);
    HAL_TIM_PWM_Start(&htim1, TIM_CHANNEL_4);
    HAL_TIM_PWM_Start(&htim3, TIM_CHANNEL_1);
    HAL_TIM_PWM_Start(&htim3, TIM_CHANNEL_2);
    HAL_TIM_PWM_Start(&htim3, TIM_CHANNEL_3);
    HAL_TIM_PWM_Start(&htim3, TIM_CHANNEL_4);
    HAL_TIM_PWM_Start(&htim2, TIM_CHANNEL_1);
    HAL_TIM_Base_Start_IT(&htim4);

    ili9341_init();

    lv_init();
    lv_disp_draw_buf_init(&disp_buf, buf_1, nullptr, DISP_BUFF_SIZE);

    static lv_disp_drv_t disp_drv;
    lv_disp_drv_init(&disp_drv);
    disp_drv.draw_buf = &disp_buf;
    disp_drv.flush_cb = ili9341_flush;
    disp_drv.hor_res = 240;
    disp_drv.ver_res = 320;
    disp_drv.rotated = LV_DISP_ROT_90;
    disp_drv.sw_rotate = 1;
    [[maybe_unused]] lv_disp_t *disp = lv_disp_drv_register(&disp_drv);

    static lv_indev_drv_t indev_drv;
    lv_indev_drv_init(&indev_drv);
    indev_drv.type = LV_INDEV_TYPE_ENCODER;
    indev_drv.read_cb = encoder_read;
    lv_indev_t *indev = lv_indev_drv_register(&indev_drv);

    ui_init(indev);

    for (;;)
    {
        lv_timer_handler_run_in_period(5);
    }
}