#include <cstring>
#include <jy61_device.h>

void jy61_device_t::handle_packet() noexcept
{
    uint8_t checksums = 0;
    for (int i = 0; i < 10; ++i)
    {
        checksums += buffer.data[i];
    }
    if (checksums != buffer.checksums)
    {
        return;
    }
    switch (buffer.kind)
    {
    case 0x51:
        __disable_irq();
        memcpy((void *)&accel_raw, (void *)(&buffer.accel), sizeof(jy61_accel_data_raw_t));
        __enable_irq();
        break;
    case 0x52:
        __disable_irq();
        memcpy((void *)&gyro_raw, (void *)(&buffer.gyro), sizeof(jy61_gyro_data_raw_t));
        __enable_irq();
        break;
    case 0x53:
        __disable_irq();
        memcpy((void *)&attitude_raw, (void *)(&buffer.attitude), sizeof(jy61_attitude_data_raw_t));
        __enable_irq();
        break;
    default:
        break;
    }
}

jy61_device_t::jy61_device_t(USART_TypeDef *usart) noexcept
    : communication_status(jy61_device_communication_status_t::PacketEnd), current_received_size(0), usart(usart)
{
}

void jy61_device_t::begin() noexcept
{
    this->usart->CR1 |= USART_CR1_RXNEIE;
    this->usart->CR3 |= USART_CR3_EIE;
}

void jy61_device_t::handle_isr() noexcept
{
    uint32_t isrflags = READ_REG(this->usart->SR);
    uint32_t cr1its = READ_REG(this->usart->CR1);
    if ((isrflags & USART_SR_RXNE) && (cr1its & USART_CR1_RXNEIE))
    {
        READ_REG(this->usart->SR);         // Read status register
        unsigned char c = this->usart->DR; // Read data register
        switch (communication_status)
        {
        case jy61_device_communication_status_t::PacketBegin:
            switch (c)
            {
            case 0x51:
            case 0x52:
            case 0x53:
                communication_status = jy61_device_communication_status_t::PacketData;
                buffer.data[current_received_size++] = c;
                break;
            default:
                communication_status = jy61_device_communication_status_t::PacketEnd;
                break;
            }
            break;
        case jy61_device_communication_status_t::PacketEnd:
            if (c == 0x55)
            {
                communication_status = jy61_device_communication_status_t::PacketBegin;
                buffer.data[0] = c;
                current_received_size = 1;
            }
            break;
        case jy61_device_communication_status_t::PacketData:
            buffer.data[current_received_size++] = c;
            if (current_received_size == sizeof(buffer))
            {
                handle_packet();
                communication_status = jy61_device_communication_status_t::PacketEnd;
            }
            break;
        }
        return;
    }
    else if (isrflags & (USART_SR_PE | USART_SR_FE | USART_SR_ORE | USART_SR_NE))
    {
        // Flags will be cleared after reading SR and then DR
        // Here we just ignore them
        // See also: Page 516, RM0368 Rev 5
        READ_REG(this->usart->SR); // Read status register
        READ_REG(this->usart->DR); // Read data register
    }
}