#include <pid_controller.h>

float pid_controller_t::compute(float error, float dt)
{
    // Compute the proportional term
    float p_term = this->Kp * error;

    // Compute the integral term
    this->error_sum += error;
    float i_term = this->Ki * this->error_sum * dt;
    if (i_term > this->integral_limit)
    {
        i_term = this->integral_limit;
        this->error_sum = this->integral_limit / this->Ki / dt;
    }
    else if (i_term < -this->integral_limit)
    {
        i_term = -this->integral_limit;
        this->error_sum = -this->integral_limit / this->Ki / dt;
    }

    // Compute the derivative term
    float d_term = this->Kd * (error - this->last_error) / dt;
    this->last_error = error;

    // Compute the total output
    float output = p_term + i_term + d_term;
    float output_rate = (output - this->last_output) / dt;
    if (output_rate > output_ramp)
    {
        output = this->last_output + this->output_ramp * dt;
    }
    else if (output_rate < -output_ramp)
    {
        output = this->last_output - this->output_ramp * dt;
    }
    this->last_output = output;

    return output;
}

void pid_controller_t::reset()
{
    this->error_sum = 0;
    this->last_error = 0;
    this->last_output = 0;
}